import os
import warnings
from pathlib import Path
from typing import List, Literal, Optional, Union

from sgqlc.operation import Operation

from ML_management.graphql import schema
from ML_management.graphql.schema import BuildArgInput, CustomImage, ImageFilterSettings, TimestampInterval
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.log_api import _open_pipe_send_request, _raise_error
from ML_management.mlmanagement.utils import calculate_hash_directory, check_exclude_include, hash_file
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.sdk.job import _get_logs
from ML_management.variables import get_log_service_url


def _base_query_image(base):
    base.name()
    base.id()
    base.description()
    base.build_args()
    base.visibility()
    base.is_active()
    base.build_job.build_object_name()
    base.build_job.status()
    base.build_job.message()


def delete_image(image_id: int) -> bool:
    """
    Marks the image as inactive. You will no longer be able to run tasks with it, but you can view information.

    Parameters
    ----------
    image_id: int
         The name of the base image.
    Returns
    -------
    None
    """
    op = Operation(schema.Mutation)
    op.delete_images(image_ids=[image_id])
    return send_graphql_request(op=op)["deleteImages"]


def get_available_images() -> List[CustomImage]:
    """Get available images for custom code job.

    Returns
    -------
    List[CustomImage]
        List of instances of the CustomImage.

    """
    op = Operation(schema.Query)
    base = op.available_images()
    _base_query_image(base)

    result = send_graphql_request(op=op, json_response=False)
    return result.available_images


def _get_image_hash(hash_data: str) -> CustomImage:
    op = Operation(schema.Query)
    base = op.get_image_with_hash(hash_data=hash_data)
    _base_query_image(base)

    result = send_graphql_request(op=op, json_response=False)
    return result.get_image_with_hash


def rebuild_image(image_id: int, build_args: Optional[dict] = None) -> CustomImage:
    """
    Rebuild user image.

    Parameters
    ----------
    image_id: int
         The id of the base image.
    build_args: Optional[dict] = None
        A dictionary of build arguments.
        If None is passed, the previous parameters will be used.
    Returns
    -------
    CustomImage
        Instance of the CustomImage class.
    """
    if build_args is not None:
        build_args = [BuildArgInput(key=key, value=str(value)) for key, value in build_args.items()]
    op = Operation(schema.Mutation)
    base = op.rebuild_image(image_id=image_id, build_args=build_args)
    _base_query_image(base)

    result = send_graphql_request(op=op, json_response=False)
    return result.rebuild_image


def log_image(
    local_path: str,
    name: str,
    description: str,
    build_args: Optional[dict] = None,
    verbose: bool = True,
    exclude: list[str | Path] | None = None,
    include: list[str | Path] | None = None,
) -> CustomImage:
    """
    Log and build user image.

    Parameters
    ----------
    local_path: str
        Path to Dockerfile or to folder with Dockerfile.
    name: str
        The name of the image.
    build_args: Optional[dict] = None
        A dictionary of build arguments.
    description: str
        Description of the image.
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.
    exclude: list[str | pathlib.Path] | None = None
        List the relative paths of the files within the logging folder that should be excluded from logging.
        For example: ["unused.py", "artifacts/modules/file.py"].
        Regular expressions are not currently supported.
        Parameters exclude and include are mutually exclusive and only one of them can be set.
    include: list[str | pathlib.Path] | None = None
        List the relative paths of only those files within the logging folder that should be included in logging.
        For example: ["model.py", "artifacts/modules/code.py"].
        Regular expressions are not currently supported.
        Parameters exclude and include are mutually exclusive and only one of them can be set.

    Returns
    -------
    CustomImage
        Instance of the CustomImage class.
    """
    local_path = os.path.abspath(local_path)
    if (os.path.isfile(local_path) and os.path.basename(local_path) != "Dockerfile") or (
        os.path.isdir(local_path) and "Dockerfile" not in os.listdir(local_path)
    ):
        raise Exception(
            "The passed local path must be the path to the Dockerfile or to the directory that contains it."
        )

    exclude, include = check_exclude_include(exclude, include)

    hash_data = calculate_hash_directory(local_path) if os.path.isdir(local_path) else hash_file(local_path)
    image = _get_image_hash(hash_data)
    if image:
        warnings.warn("Such an image already exists.")
        return image

    else:
        log_request = {"hash_data": hash_data, "name": name, "description": description}

        response = _open_pipe_send_request(
            local_path,
            log_request,
            url=get_log_service_url("log_image"),
            verbose=verbose,
            exclude=exclude,
            include=include,
        )
        _raise_error(response)
        result = response.json()
        image_id = result["image_id"]

        return rebuild_image(image_id, build_args)


def get_build_image_logs(
    name: str,
    stream: bool = True,
    file_name: Optional[str] = None,
) -> None:
    """
    Stream logs of the build job by image name.

    Parameters
    ----------
    name: str
        The name of the image.
    stream: bool = True
        Stream logs or dump all available at the moment.
    file_name: Optional[str] = None
        Name of the file where to save logs. Default: None. If None prints logs to the output.
    """
    _get_logs(
        object_type="image",
        stream=stream,
        file_name=file_name,
        params={"image_id": get_image_by_name(name).id},
    )


def get_build_image_logs_by_id(
    image_id: int,
    stream: bool = True,
    file_name: Optional[str] = None,
) -> None:
    """
    Stream logs of the build job by image id.

    Parameters
    ----------
    image_id: int
        The id of the image.
    stream: bool = True
        Stream logs or dump all available at the moment.
    file_name: Optional[str] = None
        Name of the file where to save logs. Default: None. If None prints logs to the output.
    """
    _get_logs(
        object_type="image",
        stream=stream,
        file_name=file_name,
        params={"image_id": image_id},
    )


def get_image(image_id: int) -> CustomImage:
    """
    Get image for custom code job.

    Parameters
    ----------
    image_id: int
    The id of the image.

    Returns
    -------
    CustomImage
        Instance of the CustomImage.

    """
    op = Operation(schema.Query)
    base = op.get_image(image_id=image_id)
    _base_query_image(base)

    result = send_graphql_request(op=op, json_response=False).get_image
    return result


def get_image_by_name(name: str) -> CustomImage:
    """
    Get image for custom code job.

    Parameters
    ----------
    name: str
    The name of the image.

    Returns
    -------
    CustomImage
        Instance of the CustomImage.

    """
    op = Operation(schema.Query)
    base = op.get_image_by_name(name=name)
    _base_query_image(base)

    result = send_graphql_request(op=op, json_response=False).get_image_by_name
    return result


def rename_image(image_id: int, new_name: str) -> CustomImage:
    """
    Rename image.

    Parameters
    ----------
    image_id: int
        The id of the image.
    new_name: str
        New image name.

    Returns
    -------
    CustomImage
        Instance of the CustomImage.
    """
    op = Operation(schema.Mutation)
    base = op.update_image(image_id=image_id, new_name=new_name)
    _base_query_image(base)

    image = send_graphql_request(op=op, json_response=False)
    return image.update_image


def set_image_description(image_id: int, description: str) -> CustomImage:
    """
    Set image description.

    Parameters
    ----------
    image_id: int
        The id of the image.
    description: str
        Description of an image.

    Returns
    -------
    CustomImage
        Instance of the CustomImage.
    """
    op = Operation(schema.Mutation)
    base = op.update_image(image_id=image_id, new_description=description)
    _base_query_image(base)

    image = send_graphql_request(op=op, json_response=False)
    return image.update_image


def pagination_image(
    name: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[Union[Literal["private", "public"], VisibilityOptions]] = None,
    owner_ids: Optional[list[str]] = None,
    build_end_from: Optional[int] = None,
    build_end_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[CustomImage]:
    """
    Search images.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the image.
    description: Optional[str]=None
        Description of the images.
    visibility: Optional[Union[Literal['private', 'public'], VisibilityOptions]]=None
        Visibility of image.
    owner_ids: Optional[list[str]]=None
        Ids of the image owner.
    build_end_from: Optional[int]=None
        Creation timestamp from of the model.
    build_end_to: Optional[int]=None
        Creation timestamp from of the model.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.

    Returns
    -------
    List[CustomImage]
        List of CustomImage instance with meta information.
    """
    visibility = VisibilityOptions(visibility).name if visibility else visibility
    op = Operation(schema.Query)
    base_query = op.pagination_image(
        filter_settings=ImageFilterSettings(
            image_name=name,
            description=description,
            visibility=visibility,
            owner_ids=owner_ids,
            build_end_interval=TimestampInterval(start=build_end_from, end=build_end_to),
        ),
        limit=limit,
        offset=offset,
    ).list_image
    _base_query_image(base_query)

    return send_graphql_request(op, json_response=False).pagination_image.list_image


def cancel_build_job_for_image(image_id: int) -> bool:
    """
    Cancel running or planned build job of image.

    Parameters
    ----------
    image_id: int
        Id of the image.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.cancel_build_job_for_image(image_id=image_id)
    return send_graphql_request(op)["cancelBuildJobForImage"]



"""Module for automatic jsonschema inference."""

import typing
import warnings
from copy import copy
from enum import Enum
from inspect import formatannotation, getfullargspec

from jsonschema import SchemaError
from jsonschema.validators import Draft4Validator
from pydantic import BaseModel
from pydantic.json_schema import GenerateJsonSchema

from ML_management.jsonschema_inference.jsonschema_exceptions import (
    DictKeysMustBeStringsError,
    FunctionContainsVarArgsError,
    FunctionContainsVarKwArgsError,
    InvalidSchemaError,
    InvalidStructureAnnotationError,
    NoAnnotationError,
    UnsupportedTypeError,
)

JSON_SCHEMA_DRAFT = "http://json-schema.org/draft-04/schema#"

type_map = {
    "bool": {"type": "boolean"},
    "int": {"type": "integer"},
    "float": {"type": "number"},
    "str": {"type": "string"},
}
generic_type_map = {
    "list": typing.List,  # noqa: UP006
    "dict": typing.Dict,  # noqa: UP006
    "set": typing.Set,  # noqa: UP006
    "tuple": typing.Tuple,  # noqa: UP006
}


class CustomGenerateJsonSchema(GenerateJsonSchema):
    def field_title_should_be_set(self, _) -> bool:
        return False


class SkipJsonSchema:
    """
    Annotation wrapper for system parameters to skip them in JSON schema.

    Usage example::

        def my_function(arg: Annotated[np.ndarray, SkipJsonSchema]) -> None:
            print(arg)

    """

    def __class_getitem__(cls, _type):  # for backward compatibility with older models
        return cls()


def __get_or_raise(type_name, from_optional: bool = False):
    if type_name in type_map:
        field = copy(type_map[type_name])
        if from_optional:
            field["type"] = [field["type"], "null"]
        return field
    else:
        raise UnsupportedTypeError(annotation=type_name, supported_types=list(type_map.keys()))


def __is_optional(annotation):
    repr_annotation = repr(annotation)
    if repr_annotation.startswith("typing.Optional"):
        return True
    elif (
        repr_annotation.startswith("typing.Union")
        and repr_annotation.strip("]").endswith("NoneType")
        and len(annotation.__args__) == 2
    ):
        # python 3.8 casts Optional to Union[arg, None]
        return True
    else:
        return False


def __get_json_schema_from_annotation(annotation, from_optional: bool = False, arg_name=""):
    if annotation.__class__.__name__ == "UnionType" or annotation.__class__.__name__ == "Union":
        args = list(annotation.__args__)
        if type(None) in annotation.__args__:
            from_optional = True
            args.remove(type(None))
        return __get_json_schema_from_annotation(typing.Union[tuple(args)], from_optional)

    if annotation.__module__ == "typing":
        if __is_optional(annotation):
            return __get_json_schema_from_annotation(
                annotation.__args__[0], from_optional=True
            )  # Optional[int] translates to type: [integer, null], and the field is not required
        else:
            formatted_annotation = formatannotation(annotation)

            # this is the only way to reliably get annotation name (e.g. "List")
            annotation_name = formatted_annotation.partition("[")[0]
            if not hasattr(annotation, "__args__"):
                raise InvalidStructureAnnotationError(annotation=formatted_annotation)
            if annotation_name in ["List", "Tuple"]:
                return {
                    "type": "array" if not from_optional else ["array", "null"],
                    "items": __get_json_schema_from_annotation(annotation.__args__[0]),
                }
            elif annotation_name == "Dict":
                key_annotation = annotation.__args__[0]
                if hasattr(key_annotation, "__name__"):
                    if key_annotation.__name__ != "str":
                        raise DictKeysMustBeStringsError(annotation=formatted_annotation)
                else:
                    raise DictKeysMustBeStringsError(annotation=formatted_annotation)
                return {
                    "type": "object" if not from_optional else ["object", "null"],
                    "additionalProperties": __get_json_schema_from_annotation(annotation.__args__[1]),
                }
            elif annotation_name == "Union":
                union_types = [__get_json_schema_from_annotation(ann) for ann in annotation.__args__]
                marked_as_null = False
                for type_ in union_types:
                    if type_ == {"type": "null"}:
                        marked_as_null = True
                        break
                if not marked_as_null and from_optional:
                    union_types.extend([{"type": "null"}])
                return {"anyOf": union_types}
            else:
                raise UnsupportedTypeError(
                    annotation=formatannotation(annotation),
                    supported_types=list(type_map.keys()) + ["enum.Enum", "pydantic.BaseModel"],
                )

    elif annotation.__module__ == "builtins":
        if hasattr(annotation, "__name__"):
            if annotation.__name__ in generic_type_map:
                return __get_json_schema_from_annotation(
                    generic_type_map[annotation.__name__][annotation.__args__],
                    from_optional,
                )
            return __get_or_raise(annotation.__name__, from_optional=from_optional)
    elif hasattr(annotation, "model_json_schema"):
        schema = annotation.model_json_schema(
            ref_template=f"#/properties/{arg_name}/$defs/" + "{model}",
            schema_generator=CustomGenerateJsonSchema,
        )
        schema.pop("title", None)
        schema = __convert_prefixitems_to_items(schema)
        return schema

    elif issubclass(annotation, Enum):
        return {
            "enum": [member.value for member in annotation],
            "type": "string" if not from_optional else ["string", "null"],
        }

    else:
        raise UnsupportedTypeError(
            annotation=formatannotation(annotation),
            supported_types=list(type_map.keys()),
        )


def __convert_prefixitems_to_items(schema: dict) -> dict:
    """Recursively replace 'prefixItems' with 'items' for Draft-04 compatibility."""
    if isinstance(schema, dict):
        if "prefixItems" in schema and "items" not in schema:
            schema["items"] = schema.pop("prefixItems")
        for v in schema.values():
            __convert_prefixitems_to_items(v)
    elif isinstance(schema, list):
        for item in schema:
            __convert_prefixitems_to_items(item)
    return schema


def get_function_args(spec, func):
    """Get all arguments of function."""
    # FullArgSpec(args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations)
    pos_args_defaults = {}
    if spec.defaults:
        for arg, default in zip(spec.args[-len(spec.defaults) :], spec.defaults):
            pos_args_defaults[arg] = default
    if spec.varargs:
        raise FunctionContainsVarArgsError(function_name=func.__name__)
    if spec.varkw:
        raise FunctionContainsVarKwArgsError(function_name=func.__name__)

    # annotations are not inferred from default yet, but i can do it myself from kwonlydefaults
    pos_args = spec.args if spec.args else []
    kw_only_args = spec.kwonlyargs if spec.kwonlyargs else []
    all_args = pos_args + kw_only_args
    kw_only_args_defaults = spec.kwonlydefaults if spec.kwonlydefaults else {}
    all_defaults = {**pos_args_defaults, **kw_only_args_defaults}

    return all_args, all_defaults


def serialize_value(value: typing.Any):
    """Serializes json, check types and inf."""
    if isinstance(value, Enum):
        return value.value
    elif isinstance(value, BaseModel):
        return {k: serialize_value(v) for k, v in value.model_dump().items()}
    elif isinstance(value, typing.Mapping):
        return {k: serialize_value(v) for k, v in value.items()}
    elif isinstance(value, (list, tuple, set)):
        return [serialize_value(v) for v in value]
    elif isinstance(value, (int, str, bool)) or value is None:
        return value
    elif isinstance(value, float):
        if value == float("inf"):
            raise UnsupportedTypeError(
                annotation="inf value couldn't be as default",
                supported_types=list(type_map.keys()) + ["enum.Enum", "pydantic.BaseModel"],
            )

        return value
    else:
        raise UnsupportedTypeError(
            annotation=f"default value can't have type {type(value)}",
            supported_types=list(type_map.keys()) + ["enum.Enum", "pydantic.BaseModel"],
        )


def update_schema_defaults(schema: dict, kwargs_for_init: dict) -> dict:
    """Recursively updates schema defaults."""
    schema = schema.copy()

    for key, value in schema.get("properties", {}).items():
        if key in kwargs_for_init:
            kw_val = kwargs_for_init[key]

            if isinstance(kw_val, BaseModel) and "properties" in value:
                value = update_schema_defaults(value, kw_val.model_dump())
            elif isinstance(kw_val, dict) and "properties" in value:
                value = update_schema_defaults(value, kw_val)
            else:
                value["default"] = serialize_value(kw_val)

            schema["properties"][key] = value

        elif "properties" in value:
            schema["properties"][key] = update_schema_defaults(value, kwargs_for_init.get(key, {}))

    return schema


def infer_jsonschema(func, get_object_func, **kwargs_for_init):
    """Infer jsonschema from callable by its signature."""
    from ML_management.model.model_type_to_methods_map import (
        ModelMethodName,
    )

    spec = getfullargspec(func)
    all_args, all_defaults = get_function_args(spec, func)
    schema = {
        "$schema": JSON_SCHEMA_DRAFT,
        "type": "object",
        "properties": {},
        "additionalProperties": False,
    }
    if "self" in all_args:
        all_args.remove("self")

    if func.__name__ == ModelMethodName.init.value and get_object_func:
        spec_get_object = getfullargspec(get_object_func)
        all_args_get_object, _ = get_function_args(spec_get_object, get_object_func)
        if not set(all_args_get_object) == set(all_args):
            warnings.warn(
                "get_object function arguments do not match __init__ arguments. "
                "You won't be able to pass none default arguments into __init__ function."
            )
            all_args = []
            all_defaults = []

    required = []
    for arg in all_args:
        if arg not in spec.annotations:
            raise NoAnnotationError(arg=arg)

        # skip system parameters
        if (
            typing.get_origin(spec.annotations[arg]) is typing.Annotated
            and spec.annotations[arg].__metadata__
            and spec.annotations[arg].__metadata__[0] is SkipJsonSchema
        ) or isinstance(spec.annotations[arg], SkipJsonSchema):
            continue

        schema["properties"][arg] = __get_json_schema_from_annotation(spec.annotations[arg], arg_name=arg)

        if arg in all_defaults:
            schema["properties"][arg]["default"] = all_defaults[arg]

        if arg not in all_defaults and not __is_optional(spec.annotations[arg]):
            required.append(arg)

    schema = update_schema_defaults(schema, kwargs_for_init)

    if required:
        schema["required"] = required
    try:
        Draft4Validator.check_schema(schema)
    except SchemaError as err:
        raise InvalidSchemaError(schema=schema, original_message=str(err)) from None

    try:
        ui_schema = build_ui_schema(schema)
    except:  # noqa E722
        ui_schema = None

    return {"schema": schema, "uiSchema": ui_schema}


def build_ui_schema(schema: dict, indent_step=16) -> dict:
    def collect_defs(node, acc):
        if isinstance(node, dict):
            if "$defs" in node:
                acc.update(node["$defs"])
            for v in node.values():
                collect_defs(v, acc)
        elif isinstance(node, list):
            for v in node:
                collect_defs(v, acc)

    defs = {}
    collect_defs(schema, defs)

    def resolve_ref(ref: str) -> dict:
        return defs.get(ref.split("/")[-1], {})

    def walk(node: dict, level: int) -> dict:
        ui = {}
        margin = level * indent_step

        def add_entry(key, content):
            if content:
                ui[key] = content

        # object with properties
        if node.get("type") == "object":
            for prop, prop_schema in node.get("properties", {}).items():
                entry = {}

                if level > 0:
                    entry["ui:style"] = {"marginLeft": f"{margin}px"}

                # $ref
                if "$ref" in prop_schema:
                    entry.update(walk(resolve_ref(prop_schema["$ref"]), level + 1))

                # array
                elif prop_schema.get("type") == "array":
                    items = prop_schema.get("items")

                    # tuple-style array
                    if isinstance(items, list):
                        nested = {}
                        for i, item in enumerate(items):
                            sub = walk(item, level + 1)
                            if sub:
                                nested[str(i)] = sub
                        if nested:
                            entry["items"] = nested

                    else:
                        target = resolve_ref(items["$ref"]) if isinstance(items, dict) and "$ref" in items else items
                        nested = walk(target or {}, level + 1)
                        if nested:
                            entry["items"] = nested

                # anyOf / oneOf
                elif "anyOf" in prop_schema or "oneOf" in prop_schema:
                    variants = prop_schema.get("anyOf") or prop_schema.get("oneOf")
                    merged = {}
                    for v in variants:
                        merged.update(walk(v, level + 1))
                    entry.update(merged)

                # additionalProperties
                elif prop_schema.get("type") == "object" and "additionalProperties" in prop_schema:
                    ap = prop_schema["additionalProperties"]
                    nested = walk(resolve_ref(ap["$ref"]) if "$ref" in ap else ap, level + 1)
                    if nested:
                        entry["additionalProperties"] = nested

                # plain nested object
                elif prop_schema.get("type") == "object":
                    entry.update(walk(prop_schema, level + 1))

                add_entry(prop, entry)

        return ui

    root = {"type": "object", "properties": schema.get("properties", {})}

    return walk(root, level=0)



import os
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel
from sgqlc.operation import Operation

from ML_management.graphql import schema
from ML_management.graphql.schema import CodeJob, EnvParamInput
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.git_info import get_git_info
from ML_management.mlmanagement.log_api import _open_pipe_send_request, _raise_error
from ML_management.mlmanagement.utils import calculate_hash_directory, check_exclude_include, hash_file
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.sdk.image import get_image_by_name
from ML_management.sdk.parameters import ResourcesForm
from ML_management.variables import get_log_service_url


class CodeMetaInfo(BaseModel):
    code_id: int
    experiment_name: str


def _get_code_hash(hash_code: str, visibility: VisibilityOptions) -> Optional[int]:
    op = Operation(schema.Query)
    op.get_code_with_hash(hash_code=hash_code, visibility=visibility.name)

    result = send_graphql_request(op=op, json_response=False)
    return result.get_code_with_hash


def add_custom_code_job(
    local_path: str,
    bash_commands: List[str],
    image_name: str,
    job_name: Optional[str] = None,
    exclude: list[str | Path] | None = None,
    include: list[str | Path] | None = None,
    env_variables: Optional[Dict[str, str]] = None,
    resources: Optional[ResourcesForm] = None,
    is_distributed: bool = False,
    experiment_name: str = "Default",
    experiment_visibility: Union[
        Literal[VisibilityOptions.PRIVATE, VisibilityOptions.PUBLIC], VisibilityOptions
    ] = VisibilityOptions.PRIVATE,
    visibility: Union[Literal["private", "public"], VisibilityOptions] = VisibilityOptions.PRIVATE,
    additional_system_packages: Optional[List[str]] = None,
    verbose: bool = True,
) -> CodeJob:
    """
    Create execution job from arbitrary code.

    The job created by this function can be executed on an arbitrary number of nodes greater than or equal to 1.
    The computing cluster distributes all resources equally among the allocated nodes.
    If such an equal distribution is not possible, the job will be rejected.

    Parameters
    ----------
    local_path: str
        Path to folder with code.
    bash_commands: List[str]
        Commands that will be run in the execution container.
    image_name: str
        The name of the base image on which the job will be executed.
    job_name: Optional[str] = None
        Name of the created job.
    exclude: list[str | pathlib.Path] | None = None
        List the relative paths of the files within the logging folder that should be excluded from logging.
        For example: ["unused.py", "artifacts/modules/file.py"].
        Regular expressions are not currently supported.
        Parameters exclude and include are mutually exclusive and only one of them can be set.
    include: list[str | pathlib.Path] | None = None
        List the relative paths of only those files within the logging folder that should be included in logging.
        For example: ["model.py", "artifacts/modules/code.py"].
        Regular expressions are not currently supported.
        Parameters exclude and include are mutually exclusive and only one of them can be set.
    env_variables: Optional[Dict[str, str]] = None
        Environment variables that will be set before starting the job.
    resources: ResourcesForm
        Resources required for job execution.
        They will be allocated on one node, if it is not possible, job will be rejected.
    is_distributed: bool = False
        Distributed mode.
    experiment_name: ExperimentParams
        Name of the experiment. Default: "Default"
    experiment_visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of experiment if this is new. Default: PRIVATE.
    visibility: Union[Literal['private', 'public'], VisibilityOptions]
        Visibility of this job to other users. Default: PRIVATE.
    additional_system_packages: Optional[List[str]] = None
        List of system libraries for Debian family distributions that need to be installed in the job. Default: None
    verbose: bool = True
        Whether to disable the entire progressbar wrapper.

    Returns
    -------
    CodeJob
        Instance of the Job class.
    """
    local_path = os.path.abspath(local_path)
    image_id = get_image_by_name(image_name).id
    visibility = VisibilityOptions(visibility)
    experiment_visibility = VisibilityOptions(experiment_visibility)
    if not os.path.exists(local_path):
        raise FileNotFoundError(f"Path: {local_path} does not exist.")

    exclude, include = check_exclude_include(exclude, include)

    log_request = {
        "visibility": visibility,
        "experiment_name": experiment_name,
        "experiment_visibility": experiment_visibility,
    }

    hash_code = calculate_hash_directory(local_path) if os.path.isdir(local_path) else hash_file(local_path)
    code_id = _get_code_hash(hash_code, visibility)
    experiment_params = schema.ExperimentInput(experiment_name=experiment_name, visibility=experiment_visibility.name)

    if code_id is None:
        git_info = get_git_info(local_path)
        log_request["hash_code"] = hash_code
        log_request["git_info"] = git_info.model_dump() if git_info else None
        response = _open_pipe_send_request(
            local_path,
            log_request,
            url=get_log_service_url("log_job_code"),
            verbose=verbose,
            exclude=exclude,
            include=include,
        )
        _raise_error(response)
        result = response.json()
        info = CodeMetaInfo.model_validate(result)
        code_id = info.code_id

    if resources is None:
        resources = ResourcesForm()
    resources = schema.ResourcesInput(
        cpus=resources.cpus,
        memory_per_node=resources.memory_per_node,
        gpu_number=resources.gpu_number,
        gpu_type=resources.gpu_type,
    )
    op = Operation(schema.Mutation)
    mutation = op.add_custom_code_job(
        form=schema.JobCodeParameters(
            code_id=code_id,
            bash_commands=bash_commands,
            experiment_params=experiment_params,
            visibility=VisibilityOptions(visibility).name,
            additional_system_packages=additional_system_packages,
            job_name=job_name,
            image_id=image_id,
            resources=resources,
            is_distributed=is_distributed,
            env_variables=[EnvParamInput(key=key, value=value) for key, value in env_variables.items()]
            if env_variables
            else None,
        )
    )

    mutation.name()
    mutation.id()

    job = send_graphql_request(op, json_response=False)

    return job.add_custom_code_job



from ML_management.mlmanagement.load_api import deserialize_kwargs_to_annotation


def method_params_prepare(model, method_name: str, params: dict) -> dict:
    return deserialize_kwargs_to_annotation(getattr(model, method_name), params)



import os.path
import subprocess
import warnings
from typing import Optional

from pydantic import BaseModel


class GitInfo(BaseModel):
    repo_name: str
    branch_name: str
    sha: str


def find_git_root(path: str) -> Optional[str]:
    try:
        result = subprocess.run(
            ["git", "-C", path, "rev-parse", "--show-toplevel"],
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        return None


def get_git_repo_has_changed(path: str) -> bool:
    try:
        result = subprocess.run(
            ["git", "-C", path, "diff", "HEAD"],
            capture_output=True,
            text=True,
            check=True,
        )
        return bool(result.stdout.strip())
    except subprocess.CalledProcessError:
        return False


def get_current_commit(path: str) -> Optional[str]:
    try:
        result = subprocess.run(
            ["git", "-C", path, "rev-parse", "HEAD"],
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as err:
        warnings.warn(f"Сouldn't get the meta information about git. Error: {str(err)}.")
        return None


def get_current_branch(path: str) -> Optional[str]:
    try:
        result = subprocess.run(
            ["git", "-C", path, "rev-parse", "--abbrev-ref", "HEAD"],
            capture_output=True,
            text=True,
            check=True,
        )
        branch = result.stdout.strip()
        if branch != "HEAD":
            return branch

        result = subprocess.run(
            ["git", "-C", path, "name-rev", "--name-only", "HEAD"],
            capture_output=True,
            text=True,
            check=True,
        )
        name_rev = result.stdout.strip()
        if name_rev:
            return f"{name_rev} (detached)"
        return "detached HEAD"
    except subprocess.CalledProcessError as err:
        warnings.warn(f"Сouldn't get the meta information about git. Error: {str(err)}.")
        return None


def get_git_info(local_path: str) -> Optional[GitInfo]:
    root = find_git_root(local_path)
    if not root:
        return None
    repo_name = os.path.basename(root)
    branch_name = get_current_branch(local_path)
    sha = get_current_commit(local_path)
    if not (branch_name or sha):
        return None

    if get_git_repo_has_changed(local_path):
        warnings.warn(f"Repo {repo_name} have changes.")
        branch_name += " (uncommitted changes)"

    return GitInfo(repo_name=repo_name, branch_name=branch_name, sha=sha)
